import enum
from typing import Union

import numpy as np
import numpy.linalg as la
import numpy.typing as npt

Number = Union[int, float]

EPS = 1e-5


def np_windowed(length: int, window_size: int, stride: int = 1, dilation: int = 1) -> np.ndarray:
    """
    Return indices x such that every row in array[x] is a windowed slice of the array
    Helper function which uses numpy broadcasting to work
    Adapted from: https://stackoverflow.com/a/42258242

    Args:
        length (``int``): the total length of the array to be indexed
        window_size (``int``): the size of the window to be slided across the array
        stride (``int``): the size of the "jumps" between the consecutive windows. Defaults to 1
        dilation (``int``): the distance between the elements in the window. Defaults to 1

    Returns:
        ``np.ndarray``: indices for windowing an array
    """
    base_indices: npt.NDArray[np.int_] = stride * np.arange((length - window_size) // (stride * dilation) + 1).reshape(
        -1, 1
    )
    windows: npt.NDArray[np.int_] = dilation * np.arange(window_size).reshape(1, -1)
    return windows + base_indices  # broadcasting trick


def np_anchored(length: int) -> npt.NDArray[np.int_]:
    """
    Return indices x such that every row in array[x] is a slice of the array with a first, i-th and the last element
    Helper function which uses numpy broadcasting to work

    Args:
        length (``int``): the total length of the array to be indexed in the anchored way
    Returns:
        ``np.ndarray``: indices for windowing an array

    """
    indices = list(range(1, length - 1))
    first = [0] * len(indices)
    last = [length - 1] * len(indices)
    indices = np.stack((first, indices, last), axis=-1)
    return indices


def prepare(f):
    def inner(*args, **kwargs):
        assert 1 <= len(args) <= 2
        if len(args) == 2:
            x = np.array(args[0])
            assert x.ndim == 1 and x.shape[0] > 3
            y = np.array(args[1])
            assert y.ndim == 1 and y.shape[0] > 3
            assert x.shape == y.shape
        elif len(args) == 1:
            argument = np.array(args[0]).squeeze()  # squeeze ensures all dimensions are > 1
            assert 1 <= argument.ndim <= 2
            if argument.ndim == 1:
                assert argument.shape[0] >= 3, "The input must have at least 3 points"
                y = argument
                x = np.arange(len(y))
            elif argument.ndim == 2:
                n_rows, n_cols = argument.shape
                if n_rows > n_cols:
                    argument = argument.T
                assert argument.shape[1] >= 3, "The input must have at least 3 points"
                x, y = argument[0], argument[1]
        else:
            raise ValueError("There can only be 1 or 2 positional arguments passed to the function")

        # more or less all the algorithms depend on the inputs to be sorted, at least in the x dimension
        # therefore the x and y are sorted together
        # the user should explicitly disallow sorting
        perform_sort = kwargs.pop("sort", True)
        if perform_sort and not np.all(np.diff(x) > 0):
            sorted_indices = np.argsort(x)  # sort in the ascending way
            x = x[sorted_indices]
            y = y[sorted_indices]
            assert np.all(np.diff(x))
        # all the methods should work no matter the scale of the data
        # therefore the input 2D space is transformed in [0, 1]x[0, 1] square
        x = normalize(x)
        y = normalize(y)

        # optionally smooth out the data using cubic splines (custom implementation, no external libs)
        smoothing = kwargs.pop("smoothing", 0.0)
        assert smoothing >= 0.0
        if smoothing > 0:
            x, y = cubic_spline_smoothing(x, y, smoothing)

        # knee type detection and conversion to a standard type KneeType.INCREASING_CONCAVE
        knee_type = detect_knee_type(y[0], y[1], y[-2], y[-1])
        if knee_type == KneeType.INCREASING_CONCAVE:
            return f(x, y, **kwargs)
        elif knee_type == KneeType.DECREASING_CONVEX:
            return f(x, 1 - y, **kwargs)
        elif knee_type == KneeType.INCREASING_CONVEX:
            return len(x) - f(x, 1 - y[::-1], **kwargs) - 1
        elif knee_type == KneeType.DECREASING_CONCAVE:
            return len(x) - f(x, y[::-1], **kwargs) - 1

    return inner


def normalize(x):
    """
    Helper function for normalizing the inputs.

    Normalization is an affine transformation such that the minimal element of x maps to 0, and maximal element of x
    maps to 1

    Args:
        x (``np.ndarray``): the array to be normalized

    Returns:
         ``np.ndarray``: a normalized array such that the minimum is 0 and the maximum is 1
    """
    return (x - x.min()) / (x.max() - x.min())


def get_delta_matrix(h):
    """
    Creates :math:`\\Delta` matrix. Used for cubic spline smoothing.

    Adapted from https://en.wikipedia.org/wiki/Smoothing_spline#Derivation_of_the_cubic_smoothing_spline

    [h1, h2, h3, h4] ->
    [[1/h1   -1/h1-1/h2      1/h2           0         0]
    [  0       1/h2      -1/h2-1/h3        0         0]
    [  0      0             1/h3      -1/h3-1/h4   1/h4]]

    Args:
        h (``np.ndarray``): the differences vector

    Returns:
        ``np.ndarray``: the :math:`\\Delta` matrix
    """
    assert h.ndim == 1
    n = len(h)
    dest = np.zeros((n - 1, n + 1))
    np.fill_diagonal(dest, 1 / h[:-1])
    np.fill_diagonal(dest[:, 1:], -1 / h[1:])  # heh indexing trick
    dest -= np.roll(dest, 1, axis=1)
    return dest


def get_weight_matrix(h):
    """
    Creates the weight matrix. Used for cubic spline smoothing.

    Adapted from https://en.wikipedia.org/wiki/Smoothing_spline#Derivation_of_the_cubic_smoothing_spline

    Args:
        h (``np.ndarray``): the differences vector

    Returns:
        ``np.ndarray``: the weight matrix :math:`W`

    """
    assert h.ndim == 1
    n = len(h)
    out = np.zeros((n - 1, n - 1))
    np.fill_diagonal(out, (h[:-1] + h[1:]) / 3.0)  # main diagonal
    np.fill_diagonal(out[:, 1:], h[1:] / 6.0)  # upper diagonal
    np.fill_diagonal(out[1:, :], h[1:] / 6.0)  # lower diagonal
    return out


def cubic_spline_smoothing(x, y, smoothing_factor=0):
    """
    Smoothes the :math:`y` vecetor by minimizing the second derivative.

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        y (``np.ndarray``): the :math:`y` coordinates of the points
        smoothing_factor (``float``): the cubic spline smoothing hyperparameter

    Returns:
        :obj:`tuple` of ``np.ndarray``: the :math:`x` and :math:`y` coordinates of the smoothed points
    """
    h = np.diff(x)
    delta = get_delta_matrix(h)
    weight = get_weight_matrix(h)
    # equivalent to 'delta.T @ np.inv(weight) @ delta', just both numerically more stable and faster
    matrix = delta.T @ la.solve(weight, delta)
    smoothed_y = la.solve(np.identity(len(y)) + smoothing_factor * matrix, y)
    return x, smoothed_y


def projection_distance(vertices):
    """
    Return the projection distance of the point P1 to the line through both P2 and the origin.

    Args:
        vertices (``np.ndarray``): array of shape ``(..., 2, 2)``. Coordinates of the points such that
                                    ``vertices[..., 0, :]`` are the coordinates of the point we wish to project on a
                                    line defined with the origin and ``vertices[..., 1, :]``

    Returns:
        ``np.ndarray``: one dimensional array denoting the distance the point ``vertices[..., 0, :]`` must travel to be
        projected onto the line defined by ``vertices[..., 1, :]`` and the origin

    """
    # vertices is of shape (..., 2, 2)
    assert vertices.ndim >= 2
    assert vertices.shape[-2:] == (2, 2)
    determinants = np.abs(la.det(vertices))  # this is of shape (...), the last two are not existent anymore
    vectors = vertices[..., 1, :]  # select the second row of all the matrices. this is of shape (..., 2)
    lengths = la.norm(vectors, ord=2, axis=-1)  # this is of shape (...)
    distances = determinants / (lengths + EPS)
    return distances


class KneeType(enum.Enum):
    DECREASING_CONVEX = 0
    INCREASING_CONCAVE = 1
    DECREASING_CONCAVE = 2
    INCREASING_CONVEX = 3


def detect_knee_type(y1, y2, y3, y4):
    """
    Detects the type of a knee using the first two and the last two points.

    Simplifies the problem by assuming noiseless input.

    Args:
        y1 (``float``): the :math:`y` coordinate of the first point
        y2:(``float``): the :math:`y` coordinate of the second point
        y3:(``float``): the :math:`y` coordinate of the second-to-last point
        y4:(``float``): the :math:`y` coordinate of the last point

    Returns:
        :obj:`KneeType`: the type of the knee detected
    """
    is_increasing = y3 > y2  # all the points are increasing
    is_exploding = abs(y4 - y3) > abs(y2 - y1)  # the magnitude of the increase is itself increasing #meta
    type_code = int(is_exploding) * 2 + int(is_increasing)
    return KneeType(type_code)



import numpy as np

TOLERANCE = 1e-5


def f(x, c):
    """
    The knee curve.

    This function is being fitted to the input data by tweaking the `c` parameter.

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        c (``float``): the shape parameter

    Returns:
        ``np.ndarray``: the values of the function evaluated at :math:`x` with respect to. `c`

    """
    return x * (np.exp(c) + 1) / (x * np.exp(c) + 1)


def df_dc(x, c):
    """
    The first derivative of the knee curve :math:`f(x)` with respect to :math:`x`, conditioned on :math:`c`.

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        c (``float``): the shape parameter

    Returns:
        ``np.ndarray``: values of the slopes of tangents on :math:`f(x)` evaluated at :math:`x`

    """
    t = x * np.exp(c)
    return (x - 1) * t / (t + 1) ** 2


def d2f_dc2(x, c):
    """
    The second derivative of the knee curve :math:`f(x)` with respect to :math:`x`, conditioned on :math:`c`.

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        c (``float``): the shape parameter

    Returns:
        ``np.ndarray``: values of the second derivative of :math:`f(x)` evaluated at :math:`x`

    """
    t = x * np.exp(c)
    return (x - 1) * x * np.exp(c) * (t - 1) / (t + 1) ** 3


def de_dc(y, x, c):
    """
    The first derivative of the energy function :math:`E(x)` with respect to :math:`c`.

    Used in :obj:`newton_raphson` method for the optimization of the shape parameter :math:`c`.

    Args:
        y (``np.ndarray``): the ground truth function values we wish to fit the knee curve :math:`f(x)` on
        x (``np.ndarray``): the :math:`x` coordinates of the points
        c (``float``): the shape parameter

    Returns:
        ``np.ndarray``: values of the first derivative of :math:`E(x)` evaluated at :math:`x`
    """
    return np.mean((f(x, c) - y) * df_dc(x, c))


def d2e_dc2(y, x, c):
    """
    The second derivative of the energy function :math:`E(x)` with respect to :math:`c`.

    Used in :obj:`newton_raphson` method for the optimization of the shape parameter :math:`c`.

    Args:
        y (``np.ndarray``): the ground truth function values we wish to fit the knee curve :math:`f(x)` on
        x (``np.ndarray``): the :math:`x` coordinates of the points
        c (``float``): the shape parameter

    Returns:
        ``np.ndarray``: values of the second derivative of :math:`E(x)` evaluated at :math:`x`
    """
    return np.mean(df_dc(x, c) + (f(x, c) - y) * d2f_dc2(x, c))


def newton_raphson(x, y):
    """
    The implementation of the `Newton-Raphson <https://en.wikipedia.org/wiki/Newton%27s_method>`_ optimization
    procedure.
    It fits the knee curve :math:`f(x)` to the :math:`y` s of the corresponding :math:`x` s by tweaking the shape
    parameter :math:`c` from an initial guess.

    Args:
        x (``np.ndarray``): the ground truth :math:`x` coordinates
        y (``np.ndarray``): the ground truth :math:`y` coordinates

    Returns:
        ``float``: the optimal shape parameter :math:`c` which minimizes the squared error, up to a predefined
        tolerance level

    """
    c = 0
    new_c = 3
    while abs(new_c - c) > TOLERANCE:
        c = new_c
        new_c = c - de_dc(y, x, c) / d2e_dc2(y, x, c)
    return new_c


def get_knee(c):
    """
    Calculates the position of the point of maximal curvature.

    The position depends solely on the shape parameter :math:`c`.

    Args:
        c (``float``): the shape parameter

    Returns:
        ``float``: the :math:`x` position where the knee curve :math:`f(x)` is most curved

    """
    return (np.sqrt(1 + np.exp(c)) - 1) / np.exp(c)


def c_method(x, y, **kwargs):
    """
    Implements the C-method as described in https://blog.msmetko.xyz/posts/2

    Args:
        x (``np.ndarray``): the ground truth :math:`x` coordinates
        y (``np.ndarray``): the ground truth :math:`y` coordinates
        **kwargs:``dict``, optional arguments (should be empty)

    Returns:
        ``int``: the index of the knee
    """
    assert len(kwargs) == 0
    best_c = newton_raphson(x, y)
    knee = get_knee(best_c)

    # the knee is a real number between 0 and 1 which is the best theoretical knee
    # however, that number most likely does not exist in the x array, so the closest is found
    best_knee = np.argmin(np.abs(x - knee))
    return best_knee.item()



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from collections import Counter
from functools import partial
from pathlib import Path

from knarrow import find_knee


def gte_0(value):
    x = float(value)
    assert x >= 0.0
    return x


METHODS = [
    "angle",
    "c_method",
    "distance",
    "distance_adjacent",
    "kneedle",
    "menger_anchored",
    "menger_successive",
    "ols_swiping",
]


def get_parser():
    parser = ArgumentParser(prog="knarrow", formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-m", "--method", choices=(["all"] + METHODS), default="all", help="select the knee searching method"
    )
    parser.add_argument(
        "--sort",
        action="store_true",
        help="sort the values before the knee search. By default is assumes the input is already sorted",
    )
    parser.add_argument("--smoothing", default=0.0, type=gte_0, help="cublic spline smoothing parameter")
    parser.add_argument(
        "-d", "--delimiter", default=None, help="split the values with DELIMITER. If None, split by space"
    )
    parser.add_argument(
        "-o",
        "--output",
        choices=["index", "value"],
        default="index",
        help=(
            "if output is `value`, this will return the row of the input file where the knee was detected. "
            "if output is `index`, the index of that row will be returned"
        ),
    )
    parser.add_argument("files", nargs="*", default=["-"], help="a list of files. STDIN is denoted with `-`.")
    return parser


def cli(method="all", files=None, sort=False, delimiter=None, output=None, smoothing=None):
    for filename in files:
        path = Path("/dev/stdin" if filename == "-" else filename)
        with path.open("r") as file:
            rows = list(map(str.strip, file))
            split = partial(str.split, sep=delimiter)
            values = map(split, rows)
            numbers = list(tuple(float(value) for value in row) for row in values)
            indices = list(range(len(numbers)))
            if sort:
                indices.sort(key=lambda i: numbers[i])
                key_function = (lambda x: x) if len(numbers[0]) == 1 else (lambda x: x[0])
                numbers.sort(key=key_function)

            if method == "all":
                counter = Counter([find_knee(numbers, method=m, sort=False, smoothing=smoothing) for m in METHODS])
                most_common = counter.most_common(1).pop(0)
                knee = most_common[0]
            else:
                knee = find_knee(numbers, method=method, sort=False, smoothing=smoothing)

            result = indices[knee] if output == "index" else rows[indices[knee]]
            print(path.name, result)
    return


def main():
    parser = get_parser()
    args = vars(parser.parse_args())
    exit(cli(**args))


if __name__ == "__main__":
    main()



import numpy as np


def angle(x, y, **kwargs):
    """
    Find a knee by looking at the maximum change of the angle of the line going through consecutive point pairs.

    Quite sensitive to noise, use with cubic spline smoothing.

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        y (``np.ndarray``): the :math:`y` coordinates of the points
        **kwargs: possible additional arguments (none are used)

    Returns:
        ``int``: the index of the knee
    """
    assert len(kwargs) == 0
    assert x.shape == y.shape
    d_x = np.diff(x)
    d_y = np.diff(y)
    angles = np.arctan2(d_y, d_x)
    angle_differences = np.abs(np.diff(angles))
    max_diff = angle_differences.argmax().item()
    return max_diff + 1



import numpy as np
from numpy import linalg as la


def r_squared(x, y):
    """
    Fit a linear regression to :math:`y` and return the :math:`R^2` statistical measure of a fit of a linear regression

    Args:
        x (``np.ndarray``): The :math:`x` coordinates of the points
        y (``np.ndarray``): The :math:`y` coordinates of the points

    Returns:
         ``float``: The :math:`R^2` measure of a fit. Bounded by :math:`\\left[0, 1\\right]`. The closer :math:`R^2` is
         to :math:`1`, the better the fit.

    """
    x = x.ravel()
    y = y.ravel()
    assert x.ndim == y.ndim
    assert len(x) == len(y)
    n = len(x)
    xx = np.stack((np.ones(n), x))
    w, ssr, _, _ = la.lstsq(xx.T, y, rcond=None)
    if ssr.size == 0:
        ssr = 0
    sst = np.var(y, ddof=n - 1)
    r2_1 = (1 - ssr / sst).item()
    corr = np.corrcoef(w @ xx, y)[0, 1]
    r2_2 = corr**2
    assert np.allclose(r2_1, r2_2)
    return ((r2_1 + r2_2) / 2).item()


def ols_swiping(x, y, **kwargs):
    """
    Performs OLS swiping method.

    By first selecting a pivot point, all the data points are implicitly split into two parts: the left one and the
    right one. Then, the algorithm regresses a line onto :math:`y` with respect to :math:`x` for both parts using
    ordinary least squares. If both lines fit quite well, meaning the :math:`R^2` for both left and the right fit is
    particularly high, the pivot point is declared as a knee.

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        y (``np.ndarray``): the :math:`y` coordinates of the points
        **kwargs: possible additional arguments (none are used)

    Returns:
        ``int``: the index of the knee
    """
    assert len(kwargs) == 0
    n = len(x)
    # some python magic
    # generator of all the splits, linear regressions and the sums of particular r2 coefficients
    results = ((i, r_squared(x[: i + 1], y[: i + 1]) + r_squared(x[i:], y[i:])) for i in range(1, n - 1))
    # the values are only calculated as the generator is consumed :))
    return min(results, key=lambda t: t[1])[0]



from .angle_method import angle  # noqa
from .c_method import c_method  # noqa
from .distance_method import distance, distance_adjacent  # noqa
from .kneedle import kneedle  # noqa
from .menger import menger_anchored, menger_successive  # noqa
from .ols import ols_swiping  # noqa
from .util import prepare


@prepare
def find_knee(x, y, method="menger_successive", **kwargs):
    """
    Public method for finding the knee

    Args:
        x (``np.ndarray``): the x coordinates of the points
        y (``np.ndarray``): the y coordinates of the points
        method: `str`, denotes the method to be used (default: menger_successive)
        **kwargs: possible additional arguments for the knee-finding method

    Returns (``int``): the index of the knee
    """
    assert method in [
        "angle",
        "c_method",
        "distance",
        "distance_adjacent",
        "kneedle",
        "menger_anchored",
        "menger_successive",
        "ols_swiping",
    ]
    function = globals()[method]
    return function(x, y, **kwargs)



import numpy as np

from .util import np_windowed, projection_distance


def distance(x, y, **kwargs):
    """
    Find a knee by finding a point which is most distant from the line :math:`y=x` (after normalizing the inputs)

    Fun fact: *vertical* distance of a point P from line :math:`y=x` is just a scaled version of the *orthogonal*
    distance of the same point P from line :math:`y=x`.

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        y (``np.ndarray``): the :math:`y` coordinates of the points
        **kwargs: possible additional arguments (none are actually used)

    Returns:
        ``int``: the index of the knee
    """
    assert len(kwargs) == 0
    assert x.shape == y.shape
    distances = abs(y - x)
    return np.argmax(distances).item()


def distance_adjacent(x, y, **kwargs):
    """
    Find a knee by finding a point which is most distant from the line going through the neighbouring points.

    This method is quite sensitive to noise, so only use with cubic spline smoothing.

    Note: I developed a (somewhat) fancy linear algebra implementation so this should be quite fast.

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        y (``np.ndarray``): the :math:`y` coordinates of the points
        **kwargs: possible additional arguments (none are actually used)

    Returns:
        ``int``: the index of the knee
    """
    assert len(kwargs) == 0
    indices = np_windowed(len(x), 3)
    x_windowed = x[indices]  # shape = (len(x), 3)
    y_windowed = y[indices]  # shape = (len(x), 3)
    points = np.stack((x_windowed, y_windowed), axis=-1)  # shape = (len(x), 3, 2)
    translated_points = points - points[:, [0], :]  # anchor all the triplets at the origin. The list is important!
    translated_points = translated_points[..., 1:, :]  # remove the origin po``int``, now shape = (len(x), 2, 2)
    distances = projection_distance(translated_points)
    return np.argmax(distances).item()



#!/usr/bin/env python3

from pathlib import Path

import setuptools

project_dir = Path(__file__).parent

setuptools.setup(
    name="knarrow",
    version="0.6.1",
    description="Shoot a knarrow to the knee",
    # Use UTF-8 encoding for README even on Windows by using the encoding argument.
    long_description=project_dir.joinpath("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    keywords=["python"],
    author="Marijan Smetko",
    author_email="msmetko@msmetko.xyz",
    url="https://github.com/InCogNiTo124/knarrow",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    # pip 9.0+ will inspect this field when installing to help users install a
    # compatible version of the library for their Python version.
    python_requires=">=3.7",
    # There are some peculiarities on how to include package data for source
    # distributions using setuptools. You also need to add entries for package
    # data to MANIFEST.in.
    # See https://stackoverflow.com/questions/7522250/
    include_package_data=True,
    # This file is required to inform mypy that inline type hints are used.
    #   See: https://mypy.readthedocs.io/en/stable/installed_packages.html
    package_data={"knarrow": ["py.typed"]},
    # This is a trick to avoid duplicating dependencies between both setup.py and
    # requirements.txt.
    # requirements.txt must be included in MANIFEST.in for this to work.
    # It does not work for all types of dependencies (e.g. VCS dependencies).
    # For VCS dependencies, use pip >= 19 and the PEP 508 syntax.
    #   Example: 'requests @ git+https://github.com/requests/requests.git@branch_or_tag'
    #   See: https://github.com/pypa/pip/issues/6162
    install_requires=project_dir.joinpath("requirements.txt").read_text().split("\n"),
    zip_safe=False,
    license="Apache License 2.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    entry_points={"console_scripts": ["knarrow=knarrow.cli.__main__:main"]},
)



import numpy as np
from numpy import linalg as la

from .util import np_anchored, np_windowed


def double_triangle_area(vertices):
    """
    Return twice the area of a triangle with the given vertices.

    Args:
        vertices (``np.ndarray``): :math:`3 \\times 2` matrix where every row is a new 2-D vertex :math:`(x, y)`

    Returns:
        ``np.ndarray``: :math:`1 \\times 1` array, twice the area
    """
    assert vertices.shape == (3, 2)
    double_area = abs(la.det(np.hstack((vertices, np.ones((3, 1), dtype=vertices.dtype)))))
    return double_area


def get_squared_vector_lengths(vertices):
    """
    Return the square of the lengths between neighbouring vertices

    Args:
        vertices (``np.ndarray``): array of vertices

    Returns:
        ``np.ndarray``: lenghts; the entry at ``lenghts[i]`` is a squared distance between the vertex ``i`` and ``i+1``
    """
    vector_differences = np.empty(vertices.shape)

    rolled = np.roll(vertices, -1, axis=0)
    np.subtract(rolled, vertices, out=vector_differences)
    lengths = np.einsum("ij,ij->i", vector_differences, vector_differences)
    return lengths


def get_curvature(vertices):
    """
    Calculate the Menger curvature defined by the three points

    Args:
        vertices (:obj:`np.ndarray`): :math:`3 \\times 2` matrix where every row is a new 2-D vertex :math:`(x, y)`

    Returns:
        ``np.ndarray``: the curvature, i.e. the reciprocal of the radius of the circumcircle around the vertices
    """
    area = double_triangle_area(vertices)
    value = 4 * area * area / np.prod(get_squared_vector_lengths(vertices))
    curvature = np.sqrt(value)
    return curvature


def menger_successive(x, y, **kwargs):
    """
    Find a knee using the Menger curvature on the three successive points

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        y (``np.ndarray``): the :math:`y` coordinates of the points
        **kwargs: possible additional arguments (none are used)

    Returns:
        ``int``: the index of the knee
    """
    assert len(kwargs) == 0
    assert x.shape == y.shape
    indices = np_windowed(len(x), 3)
    data_points = np.stack((x[indices], y[indices]), axis=-1)
    curve_scores = np.array([get_curvature(row) for row in data_points])
    return curve_scores.argmax().item() + 1


def menger_anchored(x, y, **kwargs):
    """
    Find a knee using the Menger curvature on the first point, last point, and varying the middle point.
    More resistant to the noise in the data than menger_successive

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        y (``np.ndarray``): the :math:`y` coordinates of the points
        **kwargs: possible additional arguments (none are used)

        Returns:
            ``int``: the index of the knee
    """
    assert len(kwargs) == 0
    assert x.shape == y.shape
    # perhaps later `menger_anchored` and `menger_successive` can be united in the future
    # since the only difference is this line
    indices = np_anchored(len(x))
    data_points = np.stack((x[indices], y[indices]), axis=-1)
    curve_scores = np.array([get_curvature(row) for row in data_points])
    return curve_scores.argmax().item() + 1



import numpy as np


def kneedle(x, y, **kwargs):
    """
    Kneedle method from https://doi.org/10.1109/ICDCSW.2011.20

    Args:
        x (``np.ndarray``): the :math:`x` coordinates of the points
        y (``np.ndarray``):  the :math:`y` coordinates of the points
        **kwargs:

    Returns:
        ``int``: the index of the knee
    """
    n = len(x)
    s = kwargs.get("S", 1.0)
    # Steps 1 and 2 from the paper are already performed
    # Step 3
    y_d = y - x

    # Step 4
    local_maxima_indices = ((y_d[1:-1] > y_d[:-2]) & (y_d[1:-1] > y_d[2:])).nonzero()[0] + 1  # len() >= 2
    if len(local_maxima_indices) == 1:
        return local_maxima_indices[0].item()
    x_lmx = x[local_maxima_indices]
    y_lmx = y_d[local_maxima_indices]
    # print(local_maxima_indices)

    # Step 5
    # The original paper states the following equation:
    # $\sum_{i=1}^{n-1}{\left(x_{sn_{i+1}}-x_{sn_i}\right)}$
    # where $x_sn$ is a vector of numbers scaled so that $min(x_sn) = 0$ and $max(x_sn) = 1$
    # If you **really** think about that, that sum will _always_ equal to 1
    t = y_lmx - s / (n - 1)

    # Step 6
    # Fancy numpy tricks:
    # 1. broadcasting to build (threshold x values) bool matrix denoting when the y_d value drops below the threshold
    # 2. clever use of np.argpartition to get the indices of the earliest True per row
    below_threshold_mask = x_lmx[:, np.newaxis] < x  # shape == (len(x_lmx), n)
    below_threshold_mask &= t[:, np.newaxis] > y_d
    # print(below_threshold_mask)
    under_threshold_ix = np.argpartition(~below_threshold_mask, 0)[:, 0]

    # filter the local maxima
    local_maxima_indices_augmented = np.append(local_maxima_indices, n)
    maxima_mask = (local_maxima_indices_augmented[:-1] < under_threshold_ix) & (
        under_threshold_ix < local_maxima_indices_augmented[1:]
    )
    final_maxima = local_maxima_indices[maxima_mask]
    return final_maxima[0].item() + 1
